library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

-------------------------------------------------------------------------------
-- package declarations for the procedures and constants.
package alupack is

  --set the default bus size
  constant bussize : integer := 4;
  
  --set up a type for a bus of size bussize
  subtype stdbus is signed (3 downto 0);
  subtype lrgbus is signed (4 downto 0);
  
  --set the integer range for a bus of size bussize + 1
  subtype medint is integer range -32 to 31;

  -- extend performs a one bit signed or signed bit extension based
  -- on the value of signex. signex=1 does a signed extension.
  procedure extend (signal inbus : in stdbus; variable outbus : out
                   lrgbus; signex : in std_logic);

  -- usadd performs signed or signed addition of two busses of size
  -- bussize. the result is a signed or signed bus of size bussize 
  -- depending on signex (signex = 1 produces a signed result). reportf 
  -- indicates if there is an underflow or overflow.
  procedure usadd (signal abus, bbus : in signed(bussize-1 downto 0);
                   signal result : out signed(bussize-1 downto 0);
                   signex : in std_logic;
                   signal reportf : out std_logic);

  -- ussub performs signed or signed subtraction (abus - bbus)
  -- of two busses of size bussize (signex=1 causes signed subtraction).
  -- reportf =1 if there is an underflow or overflow.
  procedure ussub (signal abus, bbus : in signed(bussize-1 downto 0);
                   signal result : out signed(bussize-1 downto 0);
                   signex : in std_logic;
                   signal reportf : out std_logic);
                   
end alupack;


-------------------------------------------------------------------------------
-- package body contains the procedure bodies.
package body alupack is


procedure extend (signal inbus : in stdbus; variable outbus : out   
                 lrgbus; signex : in std_logic) is
  begin
    outbus := (signex and inbus (bussize-1)) & inbus(bussize-1 downto 0);
  end;


procedure usadd (signal abus, bbus : in signed(bussize-1 downto 0);
                 signal result : out signed(bussize-1 downto 0);
                 signex : in std_logic;
                 signal reportf : out std_logic) is              
  variable tempr : medint;
  variable tempa : signed(bussize downto 0);
  variable tempb : signed(bussize downto 0);
  
  begin
    -- sign/unsign extend abus and bbus to a bus of size bussize + 1;
    extend(abus, tempa, signex);
    extend(bbus, tempb, signex);
    -- perform signed addition
    tempr := to_integer(tempa)+ to_integer(tempb);
    -- check for overflows dependent on type of addition
    if (signex = '0' and tempr > 15) then
      --overflow of signed addition
      reportf <= '1';
    elsif (signex = '1' and (tempr > 7 or tempr < -8)) then
      -- overflow or underflow of signed addition
      reportf <= '1';
    else
      reportf <= '0';
    end if;
    result <= to_signed(tempr, bussize);
  end usadd;


procedure ussub (signal abus, bbus : in signed(bussize-1 downto 0);
                 signal result : out signed(bussize-1 downto 0);
                 signex : in std_logic;
                 signal reportf : out std_logic) is                
  variable tempr : medint;
  variable tempa : signed(bussize downto 0);
  variable tempb : signed(bussize downto 0);
  
  begin
    -- sign/unsign extend abus and bbus to a bus of size bussize+1;
    extend(abus, tempa, signex);
    extend(bbus, tempb, signex);
    -- perform signed addition
    tempr := to_integer(tempa)- to_integer(tempb);
    -- check for overflows dependent on type of addition
    if (signex = '0' and tempr < 0) then
      reportf <= '1';
    elsif (signex = '1' and (tempr > 7 or tempr < -8)) then
      -- overflow or underflow of signed addition
      reportf <= '1';
    else
      reportf <= '0';
    end if;
    result <= to_signed(tempr, bussize);
  end ussub;

end alupack; -- end of package body
-------------------------------------------------------------------------------
